/* vim: set noexpandtab shiftwidth=8 cino= fo+=awc:
 ***************************************************************************
 *            logging.h
 *
 *  Copyright  2005  Matthew A. Nicholson
 *  matt@matt-land.com
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _LOGGING_H
#define _LOGGING_H

#include <ostream>
#include <streambuf>
#include <map>
#include <list>
//#include <tr1/memory>

/** @file
 *
 * This file contains VDrift's logging interface.  The interface is composed of 
 * three pieces.  The LOGGING interface provides a single access point to any 
 * LOGGERs that may be defined.  The LOGGER class encapsulates several LOGs and 
 * should be used to output log information to several different streams.  The 
 * LOG class encapsulates streams.
 *
 * The LOGGER class handles outputting data to the proper LOGs when the logging 
 * level for that LOG is high enough.  The LOG classes handle things such as 
 * placing timestamps in logs.
 */

namespace VDrift
{
	class LOGGING;
	class LOGGER;
	class LOGBUF;
	class LOG;

	// standard logging levels
	const unsigned int LOG_CRITICAL = 10;
	const unsigned int LOG_ERROR    = 20;
	const unsigned int LOG_WARN     = 30;
	const unsigned int LOG_INFO     = 40;
	const unsigned int LOG_DEBUG    = 50;
	
	// some meta logging levels
	const unsigned int LOG_NONE     = 0;
	const unsigned int LOG_DEFAULT  = LOG_ERROR;
	const unsigned int LOG_ALL      = LOG_DEBUG;
	
	
	// level setting functions to be used like: log << error << "message";
	LOGGER& critical(LOGGER&);
	LOGGER& error(LOGGER&);
	LOGGER& warn(LOGGER&);
	LOGGER& info(LOGGER&);
	LOGGER& debug(LOGGER&);

	/** The main logging singleton class.
	 *
	 * The LOGGING class holds several instances of LOGGER classes.  It 
	 * acts as a single point of access for each LOGGER, this way you don't 
	 * have to pass a handle to a LOGGER around.  The desired LOGGER is 
	 * always available through LOGGING::GetLogger().  If the requested 
	 * LOGGER does not exist, it is automatically created.
	 */
	class LOGGING
	{
		public:
			~LOGGING() { };
			static LOGGER& GetLogger(const std::string& = "");
			typedef LOG* log_t;
//			typedef std::tr1::shared_ptr< LOG > log_t;
		protected:
			// protected constructors and destructors to prevent ousiders from mucking around in here
			LOGGING();
			
			// protected copy constructors prevent copying
			explicit LOGGING(const LOGGING&);
			LOGGING& operator = (const LOGGING&);

		private:
			std::map<std::string, LOGGER> loggers;
	};
	
	/** A streambuf for LOGGERs.
	 * @internal
	 *
	 * This class is a streambuf class for LOGGERs.  It handles outputting 
	 * data to each LOG a LOGGER is managing.
	 */
	class LOGBUF : public std::streambuf
	{
		friend class LOGGER;
		public:
			typedef LOG* log_t;
//			typedef std::tr1::shared_ptr< LOG > log_t;

			explicit LOGBUF(unsigned int);
			~LOGBUF() { }

			virtual std::streambuf::int_type overflow(std::streambuf::int_type = std::streambuf::traits_type::eof());
			virtual int sync();


		private:
			unsigned int level;
			std::list< log_t > logs;

	};
	
	/** A logger object that logs messages to various streams.
	 * 
	 * The LOGGER class handles output to various logs, properly accounting 
	 * for the log level of each message and log.
	 */
	class LOGGER : public std::ostream
	{
		public:
			typedef LOG* log_t;
//			typedef std::tr1::shared_ptr< LOG > log_t;

			explicit LOGGER(unsigned int = LOG_DEFAULT);
			LOGGER(const LOGGER&);
			virtual ~LOGGER() { }

			/* TODO: Redefine endl, flush, and similar functions so  
			 * that they call the corrisponding functions in the 			   
			 * logs that are part of this LOGGER.
			 */

			// logging interface
			void AddLog(log_t);
			void AddLog(std::ostream*, unsigned int = LOG_DEFAULT);
//			void AddLog(std::tr1::shared_ptr< std::ostream >, 
//			unsigned int = LOG_DEFAULT);

			void RemoveLog(log_t);
			void RemoveLog(std::ostream*);
//			void RemoveLog(std::tr1::shared_ptr< std::ostream >);
			LOGGER& SetLevel(unsigned int);
			//unsigned int GetLevel() const { return level; }
			
			// operators
			LOGGER& operator << (LOGGER& (*func)(LOGGER&));
			
		private:
			LOGBUF buffer;
	};


	/** An actual log.
	 * This is a virtual base class although not completely abstract.
	 */
	class LOG
	{
		public:
			virtual ~LOG() { };
			virtual void put(char) = 0;
			virtual void flush() = 0;

			void SetLevel(unsigned int new_level) { level = new_level; }
			unsigned int GetLevel() const { return level; }

		protected:
			LOG(unsigned int default_level = LOG_DEFAULT) : level(default_level) { }

		private:
			unsigned int level;
	};

	/** A log pointing to a stream.
	 */
	template<typename ptr_t>
	class STREAMLOG : public LOG
	{
		public:
			STREAMLOG(ptr_t, unsigned int = LOG_DEFAULT);
			virtual ~STREAMLOG() { };

			virtual void put(char c) { stream->put(c); }
			virtual void flush() { stream->flush(); }
		private:
			ptr_t stream;
	};

	/** Construct a STREAMLOG pointing to the indicated stream with the 
	 * indicated level.
	 * @param log_stream The stream to point to.
	 * @param level The level of the stream.
	 */
	template<typename ptr_t>
	STREAMLOG<ptr_t>::STREAMLOG(ptr_t log_stream, unsigned int level) : LOG(level), stream(log_stream)
	{
	}
}

#endif
