#ifndef _PHYSICS_H
#define _PHYSICS_H

#include <stdio.h>

#include <stdlib.h>
#include <time.h>

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cmath>
#include <cassert>

#include <stdexcept>

#include <vector>
#include <list>
#include <map>

#include "btBulletCollisionCommon.h"

using namespace std;

#include "exception.h"
#include "bipointer.h"
#include "3dmath.h"
#include "vertexarray.h"

struct	MultipleRayResultCallback : public btCollisionWorld::RayResultCallback
{
	MultipleRayResultCallback(const btVector3& rayFromWorld,const btVector3& rayToWorld) : m_rayFromWorld(rayFromWorld), m_rayToWorld(rayToWorld) {}
	btVector3	m_rayFromWorld;//used to calculate hitPointWorld from hitFraction
	btVector3	m_rayToWorld;
	
	list <btCollisionWorld::LocalRayResult> results;
	
	virtual	btScalar AddSingleResult(btCollisionWorld::LocalRayResult& rayResult,bool normalInWorldSpace)
	{
		m_closestHitFraction = rayResult.m_hitFraction;
		results.push_back(rayResult);
		
		if (!normalInWorldSpace)
		{
			///need to transform normal into worldspace
			results.back().m_hitNormalLocal = rayResult.m_collisionObject->getWorldTransform().getBasis()*rayResult.m_hitNormalLocal;
		}
		//m_hitPointWorld.setInterpolate3(m_rayFromWorld,m_rayToWorld,rayResult.m_hitFraction);
		//return rayResult.m_hitFraction;
		return 1.0;
	}
};

class PHYSICSOBJECT;
class PHYSICSCOLLISIONSETTINGS;

class PHYSICSCONTACT
{
private:
	VERTEX position;
	VERTEX normal;
	float depth;
	PHYSICSOBJECT * col1, * col2;

public:
	//void SetFromODEContactGeom(dContactGeom odecontactgeom, const PHYSICSCOLLISIONSETTINGS & colsettings);
	//void SetFromBulletContact(btCollisionWorld::LocalRayResult * contact, const PHYSICSCOLLISIONSETTINGS & colsettings);
	const VERTEX GetContactPosition() const {return position;}
	const VERTEX GetContactNormal() const {return normal;}
	const float GetContactDepth() const {return depth;}
	bool operator<(const PHYSICSCONTACT& other) {return depth < other.depth;}
	PHYSICSOBJECT * GetCollidingObject1() {return col1;}
	PHYSICSOBJECT * GetCollidingObject2() {return col2;}
	void Set(const VERTEX & pos, const VERTEX & norm, float d, PHYSICSOBJECT * c1, PHYSICSOBJECT * c2)
	{
		position = pos;
		normal = norm;
		depth = d;
		col1 = c1;
		col2 = c2;
	}
};

struct PHYSICSOBJECTVERTEX
{
	float vertex[3];
};

struct PHYSICSOBJECTFACE
{
	int index[3];
};

class PHYSICSOBJECTSETTINGS
{
private:
	bool mapobject;
	bool playerobject;
	bool roadobject;
	void * objid;
	
public:
	PHYSICSOBJECTSETTINGS() : mapobject(false),playerobject(false),roadobject(false),objid(NULL) {}
	void SetMapObject(const bool newmapobject) {mapobject = newmapobject;}
	void SetPlayerObject(const bool newplayerobject) {playerobject = newplayerobject;}
	void SetRoadObject(const bool newroadobject) {roadobject = newroadobject;}
	void SetObjectID(void * newobjid) {objid = newobjid;}
	
	const bool IsMapObject() const {return mapobject;}
	const bool IsPlayerObject() const {return playerobject;}
	const bool IsRoadObject() const {return roadobject;}
	const void * const ObjID() const {return objid;}
};

class COLLISION_SPEEDUP;

class PHYSICSOBJECT : public BIPOINTER_TARGET
{
private:
	btCollisionWorld * worldid;
	btCollisionObject id;
	btBvhTriangleMeshShape * trimesh;
	btTriangleIndexVertexArray * trimesh_varray;
	
	bool loaded;
	
	PHYSICSOBJECTSETTINGS settings;
	void * parentobject;
	
	AABB bbox;
	
	COLLISION_SPEEDUP * colspeedup;

public:
	PHYSICSOBJECT() : worldid(NULL),trimesh(NULL),trimesh_varray(NULL),loaded(false),colspeedup(NULL) {}
	~PHYSICSOBJECT() {DeInit();}
	
	btCollisionObject & GetBulletObject() {return id;}
	
	void InitTrimesh(const void * vertices, int vstride, int vcount, const void * faces, int fcount, int istride, const void * normals, const PHYSICSOBJECTSETTINGS & objsettings);
	void InitTrimesh(VERTEXARRAY & varray, const PHYSICSOBJECTSETTINGS & objsettings);
	void DeInit();
	//dGeomID GetID() const {return id;}
	void SetPosition(const VERTEX & newpos);
	void SetQuaternion(const QUATERNION & newquat);
	
	void SetMapObject(const bool newmapobject) {settings.SetMapObject(newmapobject);}
	void SetPlayerObject(const bool newplayerobject) {settings.SetPlayerObject(newplayerobject);}
	void SetObjectID(void * newobjid) {settings.SetObjectID(newobjid);}
	void SetParentObject(void * newparent) {parentobject = newparent;}
	
	const bool IsMapObject() const {return settings.IsMapObject();}
	const bool IsPlayerObject() const {return settings.IsPlayerObject();}
	const bool IsRoadObject() const {return settings.IsRoadObject();}
	const void * const ObjID() const {return settings.ObjID();}
	void * ParentObject() {return parentobject;}
	
	void SetBulletWorld(btCollisionWorld * nw) {worldid = nw;}
	
	void SetCollisionSpeedup(COLLISION_SPEEDUP * colspeed)
	{
		colspeedup = colspeed;
	}
	
	inline AABB & GetBBOX() {return bbox;}
};

class PHYSICSCOLLISIONSETTINGS
{
private:
	bool mapcollide;
	bool playercollide;
	bool roadcollide;
	list <void *> exception_objectids;
	
public:
	PHYSICSCOLLISIONSETTINGS() : mapcollide(false),playercollide(false),roadcollide(false) {}
	void SetMapCollide(bool newmc) {mapcollide = newmc;}
	void SetPlayerCollide(bool newpc) {playercollide = newpc;}
	void SetRoadCollide(bool newrc) {roadcollide = newrc;}
	void SetExceptionObjectID(void * except_id) {exception_objectids.push_back(except_id);}
	const bool GetMapCollide() const {return mapcollide;}
	const bool GetPlayerCollide() const {return playercollide;}
	const bool GetRoadCollide() const {return roadcollide;}
	const list <void *> & GetExceptionObjectIDs() const {return exception_objectids;}
};

class COLLISION_NODE
{
private:
	list <PHYSICSOBJECT *> objects;
	list <COLLISION_NODE> children;
	
	AABB bbox;
	
public:
	COLLISION_NODE() {}
	
	inline const AABB & GetBBOX() const {return bbox;}
	
	void Add(PHYSICSOBJECT * object);
	void Delete(PHYSICSOBJECT * object);
	//void Query(const VERTEX & orig, const VERTEX & d, float seglen, list <PHYSICSOBJECT *> &outputlist, list <const COLLISION_NODE *> & cachenode_list_output) const;
	void Query(const AABB & querybox, list <PHYSICSOBJECT *> &outputlist) const;
	void CollapseTo(COLLISION_NODE & collapse_target);
	void DistributeObjectsToChildren(int level);
	void RemoveDuplicateObjects();
	void DebugPrint(int level, int & objectcount);
	bool Empty() {return (objects.empty() && children.empty());}
};

class COLLISION_SPEEDUP
{
private:
	vector <vector <COLLISION_NODE *> > collision_nodes;
	map <PHYSICSOBJECT *, list <pair <int, int> > > object_map; ///< speeds up deletion
	
	AABB totalsize;
	
	unsigned int GetDimensionX() {return collision_nodes.size();}
	unsigned int GetDimensionZ() {return collision_nodes[0].size();}
	
	unsigned int ConvertToGridX(float val);
	unsigned int ConvertToGridZ(float val);
	
public:
	COLLISION_SPEEDUP(int square_dimensions, float nsx, float nsz);
	~COLLISION_SPEEDUP();
	
	void Add(PHYSICSOBJECT * object);
	void Delete(PHYSICSOBJECT * object);
	//void Query(const VERTEX & orig, const VERTEX & d, float seglen, list <PHYSICSOBJECT *> &outputlist);
	void Query(const AABB & querybox, list <PHYSICSOBJECT *> &outputlist);
	void Optimize();
};

class PHYSICS
{
private:
	btCollisionWorld * id;
	btDefaultCollisionConfiguration * collisionconfig;
	btCollisionDispatcher * collisiondispatcher;
	btAxisSweep3 * collisionbroadphase;
	//dWorldID id;
	//dSpaceID playerspaceid;
	//dSpaceID mapspaceid;
	
	COLLISION_SPEEDUP colspeedup;
	
	list <PHYSICSOBJECT> objects;
	void GenerateCollisionList(const PHYSICSCOLLISIONSETTINGS & settings, list <const PHYSICSOBJECT *> & collisionlist) const;
	//void HandleCollision(const PHYSICSCOLLISIONSETTINGS & settings, MultipleRayResultCallback * contact, list <PHYSICSCONTACT> & outputcontactlist) const;
	bool PassesFilter(const PHYSICSCOLLISIONSETTINGS & settings, void * checkme) const;

public:
	PHYSICS();
	~PHYSICS() {DeInit();}
	
	void Init();
	void DeInit();
	void Tick(const double dt);
	PHYSICSOBJECT & NewPhysicsObject();
	void CollideRay(const VERTEX & position, const VERTEX & direction, const float length, list <PHYSICSCONTACT> & outputcontactlist, const PHYSICSCOLLISIONSETTINGS & settings);
	void CollideSphere(const VERTEX & position, const float size, list <PHYSICSCONTACT> & outputcontactlist, const PHYSICSCOLLISIONSETTINGS & settings);
	void CollideBox(const VERTEX & position, const QUATERNION & orientation, const VERTEX & dimensions, list <PHYSICSCONTACT> & outputcontactlist, const PHYSICSCOLLISIONSETTINGS & settings);
	//dSpaceID GetSpaceID() const {return spaceid;}
	//dSpaceID GetMapSpaceID() const {return mapspaceid;}
	//dSpaceID GetPlayerSpaceID() const {return playerspaceid;}
	void DeletePhysicsObject(PHYSICSOBJECT * objtodel);
	
	void OptimizeObjects();
};

#endif
